
import dist.DistEuclidean;

import java.util.Vector;
import java.io.RandomAccessFile;
import java.io.FileOutputStream;

import utils.IOUtils;
import utils.Utils;
import datamgr.DataReader;

/**
 * Created by IntelliJ IDEA.
 * User: dragomir
 * Date: Oct 1, 2007
 * Time: 4:09:02 PM
 * To change this template use File | Settings | File Templates.
 */
public class LightCurves_nnDists {
    private Vector lbls = new Vector();
    private Vector timeSeries = new Vector();
    private Vector strTimeSeries = new Vector();
    private double[] nnDists = null;

    public LightCurves_nnDists() throws Exception {
        String inFile = "./data/lightcurves/timeseries_9K_sorted.tts";
        String outFile = "./data/lightcurves/nndists_9K_sorted";

        readTimeSeries(inFile);
        findNNDists();
        sortNNDists();
        saveDists(outFile);
    }

    private void saveDists(String outFile) throws Exception {
        FileOutputStream fos = new FileOutputStream(outFile);
        FileOutputStream fos1 = new FileOutputStream(outFile + "_ts");
        for (int i = 0; i < nnDists.length; i++) {
            if (i % 1000 == 0) {
                System.out.println("Storing line " + i);
            }
            String line = "nn_dist = " + IOUtils.df4.format(nnDists[i]) + "\t" + lbls.get(i) + "\n";
            fos.write(line.getBytes());
            String line1 = (String) strTimeSeries.get(i);
            line1 = line1.trim();
            line1 += "\n";
            fos1.write(line1.getBytes());
        }
        fos1.close();
        fos.close();
    }

    private void sortNNDists() {
        for (int i = 0; i < nnDists.length; i++) {
            for (int j = i+1; j < nnDists.length; j++ ) {
                if (nnDists[i] > nnDists[j]) {
                    double tmpD = nnDists[i];
                    nnDists[i] = nnDists[j];
                    nnDists[j] = tmpD;

                    String tmpLbl = (String) lbls.get(i);
                    lbls.set(i, lbls.get(j));
                    lbls.set(j, tmpLbl);

                    String strTS = (String) strTimeSeries.get(i);
                    strTimeSeries.set(i, strTimeSeries.get(j));
                    strTimeSeries.set(j, strTS);

                    double[] tmpTs = (double[]) timeSeries.get(i);
                    timeSeries.set(i, timeSeries.get(j));
                    timeSeries.set(j, tmpTs);
                }
            }
        }
    }

    private void findNNDists() {
        nnDists = new double[timeSeries.size()];
        for (int i = 0; i < nnDists.length; i++) {
            nnDists[i] = Double.MAX_VALUE;
        }
        for (int i = 0; i < timeSeries.size(); i++) {
            if (i % 1000 == 0) {
                System.out.println("Computing nnDists for el#" + i);
            }
            double[] ts1 = (double[]) timeSeries.get(i);
            for (int j = 0; j < timeSeries.size(); j++) {
                if (i != j) {
                    double[] ts2 = (double[]) timeSeries.get(j);
                    updateNNDist(ts1, ts2, i);
                }
            }
        }
    }

    private void updateNNDist(double[] ts1, double[] ts2, int pos) {
        DistEuclidean dist = new DistEuclidean(true, false);
        double newDist = dist.dist(ts1, ts2, nnDists[pos]);
        if (nnDists[pos] > newDist) {
            nnDists[pos] = newDist;
        }
    }

    private void readTimeSeries(String file) throws Exception {
        DataReader reader = new DataReader(file, ' ');
        String line = "";
        int count = 0;
        while ((line = reader.readLineNoTokenize()) != null) {
            if (count % 1000 == 0) {
                System.out.println("Reading line " + count);
            }
            count++;
            lbls.add(line);
            String strts = reader.readLineNoTokenize().trim();
            strTimeSeries.add(strts);
            String[] tokens = strts.split("[\\s]");
            double[] ts = new double[1024];
            double[] zts = new double[1024];
            for (int i = 0; i < ts.length; i++) {
                ts[i] = Double.parseDouble(tokens[i]);
            }
            zts = Utils.znormalize(ts, 0, ts.length);
            timeSeries.add(zts);
        }
        reader.close();
    }

    public static void main(String[] args) throws Exception {
        new LightCurves_nnDists();
    }
}
