
import java.util.Vector;
import java.util.Hashtable;
import java.io.RandomAccessFile;
import java.io.FileOutputStream;

/**
 * Created by IntelliJ IDEA.
 * User: dragomir
 * Date: May 21, 2007
 * Time: 3:31:58 PM
 * To change this template use File | Settings | File Templates.
 */
public class ProcessLightCurves {
    private Vector lbls = new Vector();
    private Vector types = new Vector();
    private Hashtable ranksTbl = new Hashtable();
    private Vector records = new Vector();

    class record {
        int index;
        double rank;
        double rating;
        double totalrank;
        String lbl;
        String type;
    }

    ProcessLightCurves() throws Exception {
        readLbls();
        readRanks("./data/lightcurves/ogle_ranks/outliers.ceph.ppn_rank_sorted");
        readRanks("./data/lightcurves/ogle_ranks/outliers.eb.ppn_rank_sorted");
        readRanks("./data/lightcurves/ogle_ranks/outliers.rrl.ppn_rank_sorted");

        readRatings("./data/lightcurves/ogle_ranks/outliers.ceph.ppn_sorted");
        readRatings("./data/lightcurves/ogle_ranks/outliers.eb.ppn_sorted");
        readRatings("./data/lightcurves/ogle_ranks/outliers.rrl.ppn_sorted");

        createRecords();
        storeRecords("./data/lightcurves/outliers_all.txt");
        storeTopOutliers("./data/lightcurves/Topoutliers_all.txt");
        getTopDiscords("./data/lightcurves/Output_discordinds_dists.txt", "./data/lightcurves/Topdiscords_classes.txt");
    }

    private void getTopDiscords(String fileName, String outFile) throws Exception{
        Vector EB_inds = new Vector();
        Vector CEPH_inds = new Vector();
        Vector RRL_inds = new Vector();

        RandomAccessFile f = new RandomAccessFile(fileName, "r");
        String line = "";
        while ((line = f.readLine()) != null) {
            String[] tokens = line.split("[\\s,\\t]");
            int indx = Integer.parseInt(tokens[0]);
            record rec = (record) records.get(indx);
            int[] indx_rank = {indx, (int)rec.rank};
            if (rec.type.equals("EB_") && EB_inds.size() < 20) {
                EB_inds.add(indx_rank);
            } else if (rec.type.equals("CEPH_") && CEPH_inds.size() < 20) {
                CEPH_inds.add(indx_rank);
            } else if (rec.type.equals("RRL_") && RRL_inds.size() < 20) {
                RRL_inds.add(indx_rank);
            }
        }
        f.close();

        FileOutputStream fos = new FileOutputStream(outFile);
        printClass(fos, EB_inds, "EB_");
        printClass(fos, CEPH_inds, "CEPH_");
        printClass(fos, RRL_inds, "RRL_");
        fos.close();
    }

    private void printClass(FileOutputStream fos, Vector vec, String type) throws Exception {
        for (int i = 0; i < vec.size(); i++) {
            int[] indx_rank = (int[]) vec.get(i);
            String line = "" + indx_rank[0] + "\t" + indx_rank[1] +"\t" + type + "\n";
            fos.write(line.getBytes());
        }
    }

    private void storeTopOutliers(String fileName) throws Exception {
        FileOutputStream fos = new FileOutputStream(fileName);
        String line = "#index\tclassrank\trating\tglobalrank\ttype\n";
        fos.write(line.getBytes());
        for (int i = 0; i < records.size(); i++) {
            record rec = null;
            for (int j = 0; j < records.size(); j++) {
                rec = (record) records.get(j);
                if (rec.totalrank == i) {
                    break;
                }
            }
            String line1 = "" + rec.index + "\t" + (int) rec.rank + "\t" + rec.rating + "\t" + (int) rec.totalrank + "\t" + rec.type + "\t"+rec.lbl+"\n";
            fos.write(line1.getBytes());
        }
        fos.close();
    }

    private void storeRecords(String fileName) throws Exception {
        FileOutputStream fos = new FileOutputStream(fileName);
        String line = "#index\tclassrank\trating\tglobalrank\ttype\n";
        fos.write(line.getBytes());
        for (int i = 0; i < records.size(); i++) {
            record rec = (record) records.get(i);
            String line1 = "" + rec.index + "\t" + (int) rec.rank + "\t" + rec.rating + "\t" + (int) rec.totalrank + "\t" + rec.type + "\t"+rec.lbl+"\n";
            fos.write(line1.getBytes());
        }
        fos.close();
    }

    private void createRecords() {
        double[][] indx_rating = new double[lbls.size()][];
        for (int i = 0; i < lbls.size(); i++) {
            record newrec = new record();
            newrec.index = i;
            newrec.lbl = (String) lbls.get(i);
            newrec.type = (String) types.get(i);
            double[] rank_rating = (double[]) ranksTbl.get(newrec.lbl);
            newrec.rank = rank_rating[0];
            newrec.rating = rank_rating[1];
            records.add(newrec);
            double[] indx_rating_ = {i, newrec.rating};
            indx_rating[i] = indx_rating_;
        }
        computeTotalRank(indx_rating);
        for (int i = 0; i < records.size(); i++) {
            record tmprec = (record) records.get(i);
            for (int j = 0; j < indx_rating.length; j++) {
                if (indx_rating[j][0] == i) {
                    tmprec.totalrank = j;
                }
            }
        }
    }

    private void computeTotalRank(double[][] indx_rating) {
        for (int i = 0; i < indx_rating.length; i++) {
            int minpos = i;
            double minRating = indx_rating[i][1];
            for (int j = i+1; j < indx_rating.length; j++) {
                if (indx_rating[j][1] < minRating) {
                    minpos = j;
                    minRating = indx_rating[j][1];
                }
            }
            if (minpos != i) {
                double tmp_indx = indx_rating[i][0];
                double tmp_rating = indx_rating[i][1];
                indx_rating[i][0] = indx_rating[minpos][0];
                indx_rating[i][1] = indx_rating[minpos][1];
                indx_rating[minpos][0] = tmp_indx;
                indx_rating[minpos][1] = tmp_rating;
            }
        }
    }

    private void readRatings(String fileName) throws Exception {
        RandomAccessFile f = new RandomAccessFile(fileName, "r");
        String line = "";
        while ((line = f.readLine()) != null) {
            String[] tokens = line.split("[\\s,/]");
            double[] rank_rating = (double[]) ranksTbl.get(tokens[tokens.length - 1]);
            rank_rating[1] = Double.parseDouble(tokens[0]);
            ranksTbl.put(tokens[tokens.length - 1], rank_rating);
        }
        f.close();
    }

    private void readRanks(String fileName) throws Exception {
        RandomAccessFile f = new RandomAccessFile(fileName, "r");
        String line = "";
        while ((line = f.readLine()) != null) {
            String[] tokens = line.split("[\\s,/]");
            double[] rank_rating = new double[2];
            rank_rating[0] = Double.parseDouble(tokens[0]);
            ranksTbl.put(tokens[tokens.length - 1], rank_rating);
        }
        f.close();
    }

    private void readLbls() throws Exception {
        String fileName = "./data/lightcurves/TS0000_labels.txt";
        RandomAccessFile f = new RandomAccessFile(fileName, "r");
        String line = "";
        while ((line = f.readLine()) != null) {
            String[] tokens = line.split("\\s");
            lbls.add(tokens[2]);
            types.add(tokens[1]);
            System.out.println(tokens[2]);
        }
        f.close();
    }

    public static void main(String[] args) throws Exception {
        new ProcessLightCurves();
    }
}
