package datamgr;

import java.util.Vector;
import java.io.FileInputStream;

/**
 * Created by IntelliJ IDEA.
 * User: dragomir yankov
 * Date: Jun 16, 2005
 * Time: 9:15:27 AM
 * To change this template use File | Settings | File Templates.
 */
public class DataReader {
    private FileInputStream fis = null;
    private char delim = '\t';
    private byte[] buf = new byte[1024];
    private int streamRead = 0;
    private int startPos = 0;
    private String line = "";

    public DataReader(String fileName) throws Exception {
        fis = new FileInputStream(fileName);
    }

    public DataReader(String fileName, char delim) throws Exception {
        fis = new FileInputStream(fileName);
        this.delim = delim;
    }

    public void close() throws Exception {
        fis.close();
    }

    /**
     * Read a line as a single string
     */
    public String readLineNoTokenize() throws Exception {
        int pos = startPos;
        if (streamRead > 0) {
            //if there is still something in the buffer, read it first
            for ( ; streamRead > 0; streamRead--, pos++) {
                if (buf[pos] == '\n') {
                    line = new String(buf, startPos, pos - startPos);
                    String oldLine = line.trim();
                    startPos = ++pos;
                    streamRead--;
                    line = "";
                    return oldLine;
                }
            }
            if (startPos <= pos) {
                line += new String(buf, startPos, pos - startPos);
                streamRead = 0;
            }
        }
        //if the old buffer did not contain a whole line, read more data from the file
        int read = 0;
        if ((read = fis.read(buf)) > -1) {
            streamRead = read;
            //find the new line if any
            startPos = 0;
            for (pos = 0; pos < read; streamRead--, pos++) {
                if (buf[pos] == '\n') {
                    line += new String(buf, startPos, pos - startPos);
                    String oldLine = line.trim();
                    startPos = ++pos;
                    streamRead--;
                    line = "";
                    return oldLine;
                }
            }
            if (startPos < read) {
                line += new String(buf, startPos, read - startPos);
                startPos += read;
                streamRead = 0;
                return readLineNoTokenize();
            }
        }
        return null;
    }

    /**
     * Read a line and parse it into tokens 
     */
    public String[] readLine() throws Exception {
        int pos = startPos;
        if (streamRead > 0) {
            //if there is still something in the buffer, read it first
            for ( ; streamRead > 0; streamRead--, pos++) {
                if (buf[pos] == '\n') {
                    line = new String(buf, startPos, pos - startPos);
                    String[] tokens = parseLine(line.trim());
                    startPos = ++pos;
                    streamRead--;
                    line = "";
                    return tokens;
                }
            }
            if (startPos <= pos) {
                line += new String(buf, startPos, pos - startPos);
                streamRead = 0;
            }
        }
        //if the old buffer did not contain a whole line, read more data from the file
        int read = 0;
        if ((read = fis.read(buf)) > -1) {
            streamRead = read;
            //find the new line if any
            startPos = 0;
            for (pos = 0; pos < read; streamRead--, pos++) {
                if (buf[pos] == '\n') {
                    line += new String(buf, startPos, pos - startPos);
                    String[] tokens = parseLine(line.trim());
                    startPos = ++pos;
                    streamRead--;
                    line = "";
                    return tokens;
                }
            }
            if (startPos < read) {
                line += new String(buf, startPos, read - startPos);
                startPos += read;
                streamRead = 0;
                return readLine();
            }
        }
        return null;
    }

    public double[] readDoubleLine() throws Exception {
        String[] tokens = readLine();
        if (tokens == null) {
            return new double[0];
        }
        double[] res = new double[tokens.length];
        for (int i = 0; i < res.length; i++) {
            res[i] = Double.parseDouble(tokens[i]);
        }
        return res;
    }

    private String[] parseLine(String tokenLine) {
        char[] arr = tokenLine.toCharArray();
        if (arr.length == 0) {
            return null;
        }
        Vector vTokens = new Vector();
        int prevPos = 0;
        int pos = 0;
        for (; pos < arr.length; pos++) {
            if (arr[pos] == delim) {
                String nextToken = new String(arr, prevPos, pos - prevPos);
                vTokens.add(nextToken);
                while (pos < line.length() && arr[pos] == delim) {
                    pos++;
                }
                prevPos = pos;
            }
        }
        if (pos > prevPos) {
            String nextToken = (new String(arr, prevPos, pos - prevPos)).trim();
            vTokens.add(nextToken);
        }
        if (vTokens.size() == 0) {
            return null;
        }
        //can't use this because of matlab
        //return utils.Utils.toStrArray(vTokens);
        //use this instead
        return toStrArray(vTokens);
    }

    public String[] toStrArray(Vector v) {
        String[] arr = new String[v.size()];
        for (int i = 0; i < v.size(); i++) {
            arr[i] = (String) v.get(i);
        }
        return arr;
    }

    private boolean isDigit(char c) {
        return c >= '0' && c <= '9';
    }
}
