package discords;

import datamgr.DataReader;

import java.util.Vector;
import java.io.FileOutputStream;

import utils.Utils;
import utils.IOUtils;
import dist.DistEuclidean;

/**
 * Created by IntelliJ IDEA.
 * User: dragomir
 * Date: Oct 14, 2007
 * Time: 11:16:57 PM
 * To change this template use File | Settings | File Templates.
 */
public class NNDistances {
    public static char delim = '\t';
    public static boolean normalize = true;

    public void findDists(String dbDir, int numDbPages) throws Exception {
        long startTime = System.currentTimeMillis();
        System.out.println("=====================================================================================");
        //Building the main memory with all DB pages
        int offset = 0;
        Vector pages = new Vector();
        for (int i = 0; i < numDbPages; i++) {
            Vector dbPage = loadPage(dbDir, i, numDbPages);
            pages.addAll(dbPage);
        }
        long timeDuration = System.currentTimeMillis() - startTime;
        System.out.println("Computing NN finished. Time: " + (timeDuration / 1000) + " sec.");

        FileOutputStream fos = new FileOutputStream(dbDir + "nndists_allelements.txt");
        double[] dists = computeNNDists(pages);
        for (int i = 0; i < dists.length; i++) {
            fos.write((IOUtils.df4.format(dists[i]) + "\n").getBytes());
        }
        fos.close();
    }

    private double[] computeNNDists(Vector page) {
        double[] minDists = new double[page.size()];
        for (int i = 0; i < page.size(); i++) {
            if ((i % 100) == 0) {
                System.out.println("Computing NN to element# " + i);
            }
            DistEuclidean dist = new DistEuclidean(true, false);
            double minDist = Double.MAX_VALUE;
            double[] ts = (double[]) page.get(i);
            for (int j = 0; j < page.size(); j++) {
                if (j != i) {
                    double[] ts1 = (double[]) page.get(j);
                    double tmpDist = dist.dist(ts, ts1, minDist);
                    if (tmpDist < minDist) {
                        minDist = tmpDist;
                    }
                }
            }
            minDists[i] = minDist;
        }
        return minDists;
    }

    private Vector loadPage(String dir, int pageNum, int db_pages) throws Exception {
        String fileName = dir + "TS" + pageNum + ".txt";
        System.out.println("Loading page: " + fileName);
        return loadPage(fileName);
    }

    private Vector loadPage(String fileName) throws Exception {
        Vector dbPage = new Vector();
        DataReader reader = new DataReader(fileName, delim);
        String[] tokens;
        while ((tokens = reader.readLine()) != null) {
            double[] arr = new double[tokens.length];
            for (int i = 0; i < tokens.length; i++) {
                arr[i] = Double.parseDouble(tokens[i]);
            }
            if (normalize) {
                arr = Utils.znormalize(arr, 0, arr.length);
            }
            dbPage.add(arr);
        }
        reader.close();
        return dbPage;
    }

    public static void main(String[] agrs) throws Exception {
        String dir = "./data/randomwalk_1mill/";
        int numDbPages = 1;

        NNDistances dists = new NNDistances();
        dists.findDists(dir, numDbPages);
    }
}
