package discords;

import datamgr.DataReader;

import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import java.io.FileOutputStream;

import utils.Utils;
import utils.IOUtils;
import dist.DistEuclidean;

/**
 * Created by IntelliJ IDEA.
 * User: dragomir
 * Date: Oct 16, 2007
 * Time: 3:51:08 PM
 * To change this template use File | Settings | File Templates.
 */

/**
 * Similar to the Discords file, but emulates splitting the data into several computers. The maximum time for the
 * first and the second phase accross the computers is reported as overall execution time
 */
public class ParallelDiscordSimulator {
    public static char delim = '\t';
    public static boolean normalize = true;
    public static double threshold = 0.0;
    public static int computers = 0;

    public void findDiscords(String dbDir, int numDbPages, int computers, int k) throws Exception {
        Vector pageRanges = getPageRanges(numDbPages, computers);
        long maxDuration1 = 0;
        Vector vMM = new Vector();
        Vector vIndxPtrs = new Vector();
        Vector vOffsets = new Vector();
        vOffsets.add(new Integer(0));
        for (int i = 0; i < computers; i++) {
            Vector result = buildMemoryForPages(dbDir, (int[]) pageRanges.get(i));
            Vector vTmpMemory = (Vector) result.get(0);
            vMM.add(vTmpMemory);
            Vector vTmpIndxPtrs = (Vector) result.get(1);
            vIndxPtrs.add(vTmpIndxPtrs);
            Integer tmpOffset = (Integer) result.get(2);
            vOffsets.add(tmpOffset);
            Long tmpDuration = (Long) result.get(3);
            if (tmpDuration.longValue() > maxDuration1) {
                maxDuration1 = tmpDuration.longValue();
            }
        }
        System.out.println("Parallel time (phase 1)  = " + maxDuration1 + "sec.");
        mergeCandidatesLists(vMM, vIndxPtrs, vOffsets);
        System.out.println("Joined candidate set size  = " + vMM.size());

        long maxDuration2 = 0;
        Vector vMMfinal = new Vector();
        Vector vIndxPtrsfinal = new Vector();
        Vector vDistsfinal = new Vector();
        for (int i = 0; i < computers; i++) {
            int compOffset = ((Integer) vOffsets.get(i)).intValue();
            Vector result = refineDiscords(dbDir, (int[]) pageRanges.get(i), vMM, vIndxPtrs, compOffset);
            Vector vTmpMemory = (Vector) result.get(0);
            vMMfinal.add(vTmpMemory);
            Vector vTmpIndxPtrs = (Vector) result.get(1);
            vIndxPtrsfinal.add(vTmpIndxPtrs);
            Vector vTmpDists = (Vector) result.get(2);
            vDistsfinal.add(vTmpDists);
            Long tmpDuration = (Long) result.get(3);
            if (tmpDuration.longValue() > maxDuration2) {
                maxDuration2 = tmpDuration.longValue();
            }
        }
        mergeRefinedDiscords(dbDir, vMMfinal, vIndxPtrsfinal, vDistsfinal, k, maxDuration1, maxDuration2, vMM.size());
        System.out.println("Parallel time (phase 2)  = " + (maxDuration2) + "sec.");
        System.out.println("Parallel time (phase 1 & 2)  = " + (maxDuration1 + maxDuration2) + "sec.");
    }

    public Vector buildMemoryForPages(String dbDir, int[] start_endpage) throws Exception {
        Vector vMM = new Vector();
        Vector indxPtr = new Vector(); //indices of the elements in the main memory
        Vector vVolume = new Vector();
        MainMemory mm = new MainMemory();

        int offset = 0;
        long startTime = System.currentTimeMillis();
        for (int i = start_endpage[0]; i <= start_endpage[1]; i++) {
            Vector dbPage = loadPage(dbDir, i);
            mm.buildMainMemory(dbPage, threshold, vMM, vVolume, indxPtr, offset);
            offset += dbPage.size();
        }
        long endTime = System.currentTimeMillis();
        long duration = (endTime - startTime) / 1000;
        Vector result = new Vector();
        result.add(vMM);
        result.add(indxPtr);
        result.add(new Integer(offset));
        result.add(new Long(duration));
        return result;
    }

    //Do a union for all candidates lists from all computers
    public void mergeCandidatesLists(Vector vMM, Vector vIndxPtrs, Vector vOffsets) {
        Vector memory = new Vector();
        Vector inds = new Vector();
        Vector offsets = new Vector();

        int computers = vMM.size();
        int overallOffset = 0;
        for (int i = 0; i < computers; i++) {
            Vector tmpMM = (Vector) vMM.get(i);
            Vector tmpIndx = (Vector) vIndxPtrs.get(i);
            int offset = ((Integer) vOffsets.get(i)).intValue();
            overallOffset += offset;

            for (int j = 0; j < tmpMM.size(); j++) {
                memory.add(tmpMM.get(j));
                int indx = ((Integer) tmpIndx.get(j)).intValue();
                int adjIndx = overallOffset + indx;
                inds.add(new Integer(adjIndx));
            }
            offsets.add(new Integer(overallOffset));
         }
        vMM.clear();
        vIndxPtrs.clear();
        vOffsets.clear();
        vMM.addAll(memory);
        vIndxPtrs.addAll(inds);
        vOffsets.addAll(offsets);
    }

    public Vector refineDiscords(String dbDir, int[] start_endpage, Vector vMM, Vector vIndxPtrs, int compOffset) throws Exception {
        Vector nnDists = new Vector();
        for (int i = 0; i < vMM.size(); i++) {
            nnDists.add(new Double(Double.MAX_VALUE));
        }
        Vector vVolume = new Vector();
        int offset = compOffset;
        Vector memory = copyMemory(vMM);
        Vector inds = copyInds(vIndxPtrs);
        long startTime = System.currentTimeMillis();
        for (int i = start_endpage[0]; i <= start_endpage[1]; i++) {
            Vector dbPage = loadPage(dbDir, i);
            findDiscords(dbPage, memory, nnDists, inds, vVolume, threshold, offset);
            offset += dbPage.size();
        }
        long endTime = System.currentTimeMillis();
        long duration = (endTime - startTime) / 1000;
        Vector result = new Vector();
        result.add(memory);
        result.add(inds);
        result.add(nnDists);
        result.add(new Long(duration));
        return result;
    }

    //Do an intersection of the discords from all computers
    public void mergeRefinedDiscords(String dbDir, Vector vMM, Vector vIndxPtrs, Vector vDists, int k, long duration1, long duration2, int candidates) {
        int computers = vMM.size();
        Hashtable discords = new Hashtable();
        for (int i = 0; i < computers; i++) {
            Vector tmpMem = (Vector) vMM.get(i);
            Vector tmpInds = (Vector) vIndxPtrs.get(i);
            Vector tmpDists = (Vector) vDists.get(i);
            for (int j = 0; j < tmpMem.size(); j++) {
                Integer ptr = (Integer) tmpInds.get(j);
                Vector disc = (Vector) discords.get(ptr);
                if (disc == null && i > 0) {  //a discord in some of the computers that was not found on some computer after that
                    continue;
                }
                if (disc == null) { //all discords from the first computer will be added, need to check whether they are present in the other
                                    //computers too, if not remove them
                    disc = new Vector();
                    disc.add(tmpMem.get(j));
                    Vector dists = new Vector();
                    disc.add(dists);
                }
                Vector dists = (Vector) disc.get(1);
                dists.add(tmpDists.get(j));
                discords.put(ptr, disc);
            }
        }
        Vector vdisc = (Vector) discords.get(new Integer(1251));
        Vector finalDiscords = new Vector();
        for (Enumeration keys = discords.keys(); keys.hasMoreElements();) {
            Integer ptr = (Integer) keys.nextElement();
            Vector disc = (Vector) discords.get(ptr);
            Vector dists = (Vector) disc.get(1);
            if (dists.size() == computers) { //the discord was a true discord accross all computers
                Double minDist = getMinDist(dists);
                disc.set(1, ptr);
                disc.add(minDist);
                finalDiscords.add(disc);
            }
            discords.remove(ptr);
        }
        vMM.clear();
        vIndxPtrs.clear();
        vDists.clear();

        sortDiscords(finalDiscords);
        logDiscords(dbDir, finalDiscords, k, duration1, duration2, candidates);
    }

    private void findDiscords(Vector DB, Vector MM, Vector nnDists, Vector indxPtr, Vector vVolume, double threshold, int offset) {
        DistEuclidean dist = new DistEuclidean(true, false);
        for (int i = 0; i < MM.size(); ) {
            boolean toRemove = false;
            int elIndx = ((Integer) indxPtr.get(i)).intValue();
            double currMinNNDist = ((Double) nnDists.get(i)).doubleValue();
            for (int j = 0; j < DB.size(); j++) {
                int pos = offset + j;
                if (pos == elIndx) {
                    continue;
                }
                double d = dist.dist((double[]) MM.get(i), (double[]) DB.get(j), currMinNNDist);
                if (d < threshold) {
                    toRemove = true;
                    break;
                }
                if (d < currMinNNDist) {
                    nnDists.set(i, new Double(d));
                    currMinNNDist = d;
                }
            }
            if (toRemove) {
                MM.remove(i);
                nnDists.remove(i);
                indxPtr.remove(i);
            } else {
                i++;
            }
        }
        vVolume.add(new Integer(MM.size()));
    }

    private Vector copyMemory(Vector MM) {
        Vector memory = new Vector();
        for (int i = 0 ; i < MM.size(); i++) {
            double[] ts = (double[]) MM.get(i);
            double[] newts = new double[ts.length];
            System.arraycopy(ts, 0, newts, 0, newts.length);
            memory.add(newts);
        }
        return memory;
    }

    private Vector copyInds(Vector vIndsPtrs) {
        Vector inds = new Vector();
        for (int i = 0; i < vIndsPtrs.size(); i++) {
            int ptr = ((Integer) vIndsPtrs.get(i)).intValue();
            inds.add(new Integer(ptr));
        }
        return inds;
    }

    private Vector loadPage(String dir, int pageNum) throws Exception {
        String fileName = dir + "TS" + pageNum + ".txt";
        System.out.println("Loading page: " + fileName);
        return loadPage(fileName);
    }

    private Vector loadPage(String fileName) throws Exception {
        Vector dbPage = new Vector();
        DataReader reader = new DataReader(fileName, delim);
        String[] tokens;
        while ((tokens = reader.readLine()) != null) {
            double[] arr = new double[tokens.length];
            for (int i = 0; i < tokens.length; i++) {
                arr[i] = Double.parseDouble(tokens[i]);
            }
            if (normalize) {
                arr = Utils.znormalize(arr, 0, arr.length);
            }
            dbPage.add(arr);
        }
        reader.close();
        return dbPage;
    }

    public Vector getPageRanges(int numDbPages, int computers) {
        Vector ranges = new Vector();
        int pagesPerComp = numDbPages / computers;
        int remainingPages = numDbPages - (pagesPerComp * computers);
        int areRemainingPages = (remainingPages > 0) ? 1 : 0;
        int[] start_end;
        int startPage = 0;
        int endPage = pagesPerComp+ areRemainingPages - 1;
        for (int i = 0; i < remainingPages; i++) {
            start_end = new int[] {startPage, endPage};
            ranges.add(start_end);
            startPage = endPage + 1;
            endPage = startPage + pagesPerComp;
        }
        endPage -= areRemainingPages;
        for (int i = remainingPages; i < computers; i++) {
            start_end = new int[] {startPage, endPage};
            ranges.add(start_end);
            startPage += pagesPerComp;
            endPage = startPage + pagesPerComp - 1;
        }
        return ranges;
    }

    private void logDiscords(String dir, Vector discords, int k, long duration1, long duration2, int candidates) {
        try {
            FileOutputStream fos = new FileOutputStream(dir + "parallel_discords" + "_" + computers+"computers");
            FileOutputStream fos1 = new FileOutputStream(dir + "parallel_indices_dist_volume"+ "_"+computers+"computers");
            int min = (k < discords.size()) ? k : discords.size();
            fos1.write("The first index is 0\n".getBytes());
            for (int i = 0; i < min; i++) {
                String line = "";
                Vector disc = (Vector) discords.get(i);
                double[] ts = (double[]) disc.get(0);
                for (int j = 0; j < ts.length; j++) {
                    line += (IOUtils.df4.format(ts[j]) + "\t");
                }
                line = line.trim();
                line += "\n";
                fos.write(line.getBytes());
                String line1 = ("Index= " + ((Integer) disc.get(1)).intValue() + "\tnnDist= " + (IOUtils.df4.format(((Double) disc.get(2)).doubleValue())));
                line1 += "\n";
                fos1.write(line1.getBytes());
            }
            fos1.write(("Discord candidates (union all computers): " + (candidates) + "\n").getBytes());
            fos1.write(("Discords : " + (discords.size()) + "\n").getBytes());
            fos1.write(("Time Phase1 (max accross all computers): " + duration1 + "\n").getBytes());
            fos1.write(("Time Phase2 (max accross all computers): " + duration2 + "\n").getBytes());
            fos1.write(("Time Total (Phase1+Phase2): " + (duration1 + duration2) + "\n").getBytes());

            fos1.close();
            fos.close();
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    private void sortDiscords(Vector discords) {
        for (int i = 0; i < discords.size(); i++) {
            Vector currDisc = (Vector) discords.get(i);
            double dist = ((Double) currDisc.get(2)).doubleValue();
            for (int j = i + 1; j < discords.size(); j++) {
                Vector currDisc1 = (Vector) discords.get(j);
                double dist1 = ((Double) currDisc1.get(2)).doubleValue();
                if (dist1 > dist) {
                    discords.set(j, currDisc);
                    discords.set(i, currDisc1);
                    currDisc = (Vector) discords.get(i);
                    dist = dist1;
                }
            }
        }
    }

    private Double getMinDist(Vector dists) {
        double minDist = Double.MAX_VALUE;
        for (int i = 0; i < dists.size(); i++) {
            double tmpDist = ((Double) dists.get(i)).doubleValue();
            if (tmpDist < minDist) {
                minDist = tmpDist;
            }
        }
        return new Double(minDist);
    }

    public static void main(String[] args) throws Exception {
//        String dir = "./data/lightcurves/";
//        int numDbPages = 10;
//        threshold = 10;
//        int numDiscords = 10;
        String dir = "./data/randomwalk_1mill/";
        int numDbPages = 100;
        threshold = 21.45;
        int numDiscords = 10;
        computers = 8;

        normalize = false;

        ParallelDiscordSimulator simulator = new ParallelDiscordSimulator();
        simulator.findDiscords(dir, numDbPages, computers, numDiscords);
    }
}
